
import asyncio
import copy
import datetime
import hashlib
import logging
import socket
import sys
import time

from qtoggleserver import system
from qtoggleserver import version
from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core.typing import Attributes, AttributeDefinitions, GenericJSONDict
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils.cmd import run_set_cmd

from . import events as device_events


logger = logging.getLogger(__name__)

ATTRDEFS = {
    'name': {
        'type': 'string',
        'modifiable': True,
        'persisted': True,
        'min': 1,
        'max': 32,
        'pattern': r'^[_a-zA-Z][_a-zA-Z0-9-]{0,31}$',
        'standard': True
    },
    'display_name': {
        'type': 'string',
        'modifiable': True,
        'max': 64,
        'standard': True
    },
    'version': {
        'type': 'string',
        'standard': True
    },
    'api_version': {
        'type': 'string',
        'standard': True
    },
    'vendor': {
        'type': 'string',
        'standard': True
    },
    'admin_password': {
        'type': 'string',
        'modifiable': True,
        'max': 32,
        'standard': True
    },
    'normal_password': {
        'type': 'string',
        'modifiable': True,
        'max': 32,
        'standard': True
    },
    'viewonly_password': {
        'type': 'string',
        'modifiable': True,
        'max': 32,
        'standard': True
    },
    'flags': {
        'type': ['string'],
        'standard': True
    },
    'virtual_ports': {
        'type': 'number',
        'enabled': lambda: bool(settings.core.virtual_ports),
        'standard': True
    },
    'uptime': {
        'type': 'number',
        'standard': True
    },
    'date': {
        'type': 'string',
        'pattern': r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$',
        'modifiable': lambda: system.date.has_set_date_support(),
        'persisted': False,
        'standard': False  # Having standard False here enables exposing of definition (needed for non-modifiable)
    },
    'timezone': {
        'type': 'string',
        'modifiable': True,
        'persisted': False,
        'choices': [{'value': zone} for zone in system.date.get_timezones()],
        'enabled': lambda: system.date.has_timezone_support(),
        'standard': False  # Having standard False here enables exposing of definition (needed for choices)
    },
    'wifi_ssid': {
        'type': 'string',
        'max': 32,
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'wifi_key': {
        'type': 'string',
        'max': 64,
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'wifi_bssid': {
        'type': 'string',
        'pattern': r'^([a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2})?$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'wifi_bssid_current': {
        'type': 'string',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'wifi_signal_strength': {
        'type': 'number',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'ip_address': {
        'type': 'string',
        'pattern': r'^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_netmask': {
        'type': 'number',
        'min': 0,
        'max': 31,
        'integer': True,
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_gateway': {
        'type': 'string',
        'pattern': r'^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_dns': {
        'type': 'string',
        'pattern': r'^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_address_current': {
        'type': 'string',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_netmask_current': {
        'type': 'number',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_gateway_current': {
        'type': 'string',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_dns_current': {
        'type': 'string',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    }
}

EMPTY_PASSWORD_HASH = hashlib.sha256(b'').hexdigest()

WIFI_RSSI_EXCELLENT = -50
WIFI_RSSI_GOOD = -60
WIFI_RSSI_FAIR = -70

NETWORK_ATTRS_WATCH_INTERVAL = 10


name = socket.gethostname()
display_name = ''
admin_password_hash = None
normal_password_hash = None
viewonly_password_hash = None

_schema = None
_attrdefs = None
_attrs_watch_task = None


class DeviceAttributeError(Exception):
    pass


def get_attrdefs() -> AttributeDefinitions:
    global _attrdefs

    if _attrdefs is None:
        logger.debug('initializing attribute definitions')
        _attrdefs = copy.deepcopy(ATTRDEFS)

        # Transform all callable values into corresponding results
        for n, attrdef in _attrdefs.items():
            for k, v in attrdef.items():
                if callable(v):
                    attrdef[k] = v()

    return _attrdefs


def get_schema() -> GenericJSONDict:
    global _schema

    if not _schema:
        _schema = {
            'type': 'object',
            'properties': {},
            'additionalProperties': False
        }

        for name, attrdef in get_attrdefs().items():
            if not attrdef.get('modifiable'):
                continue

            attr_schema = dict(attrdef)

            enabled = attr_schema.pop('enabled', True)
            if not enabled:
                continue

            if attr_schema['type'] == 'string':
                if 'min' in attr_schema:
                    attr_schema['minLength'] = attr_schema.pop('min')

                if 'max' in attr_schema:
                    attr_schema['maxLength'] = attr_schema.pop('max')

            elif attr_schema['type'] == 'number':
                if attr_schema.get('integer'):
                    attr_schema['type'] = 'integer'

                if 'min' in attr_schema:
                    attr_schema['minimum'] = attr_schema.pop('min')

                if 'max' in attr_schema:
                    attr_schema['maximum'] = attr_schema.pop('max')

            if 'choices' in attrdef:
                attr_schema['enum'] = [c['value'] for c in attr_schema.pop('choices')]

            attr_schema.pop('persisted', None)
            attr_schema.pop('modifiable', None)
            attr_schema.pop('standard', None)

            _schema['properties'][name] = attr_schema

    return _schema


def get_attrs() -> Attributes:
    attrs = {
        'name': name,
        'display_name': display_name,
        'version': version.VERSION,
        'api_version': core_api.API_VERSION,
        'vendor': version.VENDOR,
        'uptime': system.uptime(),

        # Never disclose passwords
        'admin_password': '',
        'normal_password': '',
        'viewonly_password': ''
    }

    flags = ['expressions']
    if settings.system.fwupdate_driver:
        flags.append('firmware')

    if settings.core.listen_support:
        flags.append('listen')

    if settings.slaves.enabled:
        flags.append('master')

    if settings.reverse.enabled:
        flags.append('reverse')

    if settings.core.sequences_support:
        flags.append('sequences')

    if settings.core.ssl_support:
        flags.append('ssl')

    if settings.webhooks.enabled:
        flags.append('webhooks')

    attrs['flags'] = flags

    if settings.core.virtual_ports:
        attrs['virtual_ports'] = settings.core.virtual_ports

    if system.date.has_real_date_time():
        attrs['date'] = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')

    if system.date.has_timezone_support():
        attrs['timezone'] = system.date.get_timezone()

    if system.net.has_wifi_support():
        wifi_config = system.net.get_wifi_config()
        attrs['wifi_ssid'] = wifi_config['ssid']
        attrs['wifi_key'] = wifi_config['psk']
        attrs['wifi_bssid'] = wifi_config['bssid']
        attrs['wifi_bssid_current'] = wifi_config['bssid_current']

        strength = -1
        rssi = wifi_config['rssi_current']
        if rssi:
            rssi = int(rssi)
            if rssi >= WIFI_RSSI_EXCELLENT:
                strength = 3

            elif rssi >= WIFI_RSSI_GOOD:
                strength = 2

            elif rssi >= WIFI_RSSI_FAIR:
                strength = 1

            else:
                strength = 0

        attrs['wifi_signal_strength'] = strength

    if system.net.has_ip_support():
        ip_config = system.net.get_ip_config()
        attrs['ip_address'] = ip_config['address']
        attrs['ip_netmask'] = int(ip_config['netmask'] or 0)
        attrs['ip_gateway'] = ip_config['gateway']
        attrs['ip_dns'] = ip_config['dns']

        if 'address_current' in ip_config:
            attrs['ip_address_current'] = ip_config['address_current']
        if 'netmask_current' in ip_config:
            attrs['ip_netmask_current'] = int(ip_config['netmask_current'] or 0)
        if 'gateway_current' in ip_config:
            attrs['ip_gateway_current'] = ip_config['gateway_current']
        if 'dns_current' in ip_config:
            attrs['ip_dns_current'] = ip_config['dns_current']

    return attrs


def set_attrs(attrs: Attributes) -> bool:
    core_device_attrs = sys.modules[__name__]

    reboot_required = False
    attrdefs = get_attrdefs()

    for name, value in attrs.items():
        # A few attributes may carry sensitive information, so treat them separately and do not log their values
        if name.count('password') or name == 'wifi_key':
            logger.debug('setting device attribute %s', name)

        else:
            logger.debug('setting device attribute %s = %s', name, json_utils.dumps(value))

        attrdef = attrdefs[name]

        if not attrdef.get('modifiable'):
            raise DeviceAttributeError(f'attribute not modifiable: {name}')

        # Treat passwords separately, as they are not persisted as given, but hashed first
        if name.endswith('_password') and hasattr(core_device_attrs, name + '_hash'):
            # Call password set command, if available
            if settings.core.passwords.set_cmd:
                run_set_cmd(
                    settings.core.passwords.set_cmd,
                    cmd_name='password',
                    log_values=False,
                    username=name[:-9],
                    password=value
                )

            value = hashlib.sha256(value.encode()).hexdigest()
            name += '_hash'

            setattr(core_device_attrs, name, value)
            continue

        persisted = attrdef.get('persisted', attrdef.get('modifiable'))
        if persisted:
            setattr(core_device_attrs, name, value)

        if name == 'name' and settings.core.device_name.set_cmd:
            run_set_cmd(settings.core.device_name.set_cmd, cmd_name='device name', name=value)

        elif name == 'date' and system.date.has_set_date_support():
            try:
                date = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')

            except ValueError:
                raise DeviceAttributeError(f'invalid field: {name}')

            system.date.set_date(date)

        elif name == 'timezone' and system.date.has_timezone_support():
            system.date.set_timezone(value)

        elif name in ('wifi_ssid', 'wifi_key', 'wifi_bssid') and system.net.has_wifi_support():
            wifi_config = system.net.get_wifi_config()
            k = name[5:]
            k = {
                'key': 'psk'
            }.get(k, k)
            wifi_config[k] = value
            wifi_config = {k: v for k, v in wifi_config.items() if not k.endswith('_current')}

            system.net.set_wifi_config(**wifi_config)
            reboot_required = True

        elif name in ('ip_address', 'ip_netmask', 'ip_gateway', 'ip_dns') and system.net.has_ip_support():
            ip_config = system.net.get_ip_config()

            k = name[3:]
            ip_config[k] = value
            ip_config = {k: v for k, v in ip_config.items() if not k.endswith('_current')}
            ip_config['netmask'] = str(ip_config['netmask'])

            system.net.set_ip_config(**ip_config)
            reboot_required = True

    return reboot_required


def to_json() -> GenericJSONDict:
    attrdefs = copy.deepcopy(get_attrdefs())
    filtered_attrdefs = {}
    for n, attrdef in attrdefs.items():
        if attrdef.pop('standard', False):
            continue

        enabled = attrdef.pop('enabled', True)
        if not enabled:
            continue

        attrdef.pop('persisted', None)
        attrdef.pop('pattern', None)

        filtered_attrdefs[n] = attrdef

    result = dict(get_attrs())
    result['definitions'] = filtered_attrdefs

    return result


def _check_net_data_changed(data: dict) -> bool:
    changed = False

    if system.net.has_wifi_support():
        wifi_config = system.net.get_wifi_config()
        old_wifi_config = data.get('wifi_config')
        if old_wifi_config != wifi_config:
            data['wifi_config'] = wifi_config
            changed = True

    if system.net.has_ip_support():
        ip_config = system.net.get_ip_config()
        old_ip_config = data.get('ip_config')
        if old_ip_config != ip_config:
            data['ip_config'] = ip_config
            changed = True

    return changed


async def _attrs_watch_loop() -> None:
    last_net_time = time.time()
    last_net_data = {}

    try:
        while True:
            now = time.time()
            changed = False
            if now - last_net_time >= NETWORK_ATTRS_WATCH_INTERVAL:
                try:
                    if _check_net_data_changed(last_net_data):
                        logger.debug('network attributes data changed')
                        changed = True

                except Exception as e:
                    logger.error('network attributes data check failed: %s', e, exc_info=True)

                last_net_time = now

            if changed:
                device_events.trigger_update()

            await asyncio.sleep(1)

    except asyncio.CancelledError:
        logger.debug('attributes watch task cancelled')


async def init() -> None:
    global _attrs_watch_task

    logger.debug('starting attributes watch task')
    _attrs_watch_task = asyncio.create_task(_attrs_watch_loop())


async def cleanup() -> None:
    logger.debug('stopping attributes watch task')
    _attrs_watch_task.cancel()
    await _attrs_watch_task




import datetime
import json

from typing import Any, Callable, Union

import jsonpointer


JSON_CONTENT_TYPE = 'application/json; charset=utf-8'


def _resolve_refs_rec(obj: Any, root_obj: Any) -> Any:
    if isinstance(obj, dict):
        if len(obj.keys()) == 1 and list(obj.keys())[0] == '$ref':
            ref = list(obj.values())[0]
            ref = ref[1:]  # Skip starting hash
            return jsonpointer.resolve_pointer(root_obj, ref)

        for k, v in obj.items():
            obj[k] = _resolve_refs_rec(v, root_obj)

    elif isinstance(obj, list):
        for i, e in enumerate(obj):
            obj[i] = _resolve_refs_rec(e, root_obj)

    return obj


def _make_json_encoder(date_format: str = '%Y-%m-%d', datetime_format: str = '%Y-%m-%dT%H:%M:%SZ') -> Callable:
    def encode_default_json(obj: Any) -> str:
        if isinstance(obj, datetime.datetime):
            return obj.strftime(datetime_format)

        elif isinstance(obj, datetime.date):
            return obj.strftime(date_format)

        elif isinstance(obj, (set, tuple)):
            return json.dumps(list(obj), default=encode_default_json)

    return encode_default_json


def dumps(obj: Any) -> str:
    if isinstance(obj, str):
        return '"' + obj + '"'

    elif isinstance(obj, bool):
        return ['false', 'true'][obj]

    elif isinstance(obj, (int, float)):
        return str(obj)

    elif obj is None:
        return 'null'

    else:
        return json.dumps(obj, default=_make_json_encoder())


def loads(s: Union[str, bytes], resolve_refs: bool = False) -> Any:
    obj = json.loads(s)

    if resolve_refs:
        obj = _resolve_refs_rec(obj, root_obj=obj)

    return obj




from typing import Optional, Set

from qtoggleserver.core.typing import PortValue as CorePortValue

from .base import Expression
from .exceptions import IncompleteExpression

# Import core.ports after defining Expression, because core.ports.BasePort depends on Expression.
from qtoggleserver.core import ports as core_ports


class PortValue(Expression):
    def __init__(self, port_id: str) -> None:
        self.port_id: str = port_id

    def __str__(self) -> str:
        return f'${self.port_id}'

    def get_deps(self) -> Set[str]:
        return {f'${self.port_id}'}

    def get_port(self) -> core_ports.BasePort:
        return core_ports.get(self.port_id)

    def eval(self) -> CorePortValue:
        port = self.get_port()
        if not port:
            raise IncompleteExpression(f'Unknown port {self.port_id}')

        if not port.is_enabled():
            raise IncompleteExpression(f'{port} is disabled')

        value = port.get_value()
        if value is None:
            raise IncompleteExpression(f'Value of port {port} is undefined')

        return float(value)

    @staticmethod
    def parse(self_port_id: Optional[str], sexpression: str) -> Expression:
        sexpression = sexpression.strip()
        port_id = sexpression.strip('$')

        if port_id:
            return PortValue(port_id)

        else:
            return SelfPortValue(self_port_id)


class SelfPortValue(PortValue):
    def __str__(self) -> str:
        return '$'




from qtoggleserver import version
from qtoggleserver.conf import settings

from . import constants


def make_context() -> dict:
    return {
        'slaves_enabled': settings.slaves.enabled,

        'name': constants.FRONTEND_APP_NAME,
        'display_name': constants.FRONTEND_APP_DISPLAY_NAME,
        'version': version.VERSION,
        'debug': settings.frontend.debug,
        'static_url': ['static', 'static/app'][settings.frontend.debug],
        'navigation_base_prefix': '/' + constants.FRONTEND_URL_PREFIX,
        'themes': ['dark', 'light']
    }




import importlib
import logging
import os

from typing import List, Optional, Union

from jinja2 import Environment, FileSystemLoader, select_autoescape

from .constants import FRONTEND_DIR, FRONTEND_DIR_DEBUG


_env: Optional[Environment] = None
logger = logging.getLogger(__name__)


class J2TemplateMixin:
    def render(self, template_name: str, **context) -> None:
        env = get_env()
        template = env.get_template(template_name)
        template_str = template.render(**context)

        self.finish(template_str)


class NamespaceLoader(FileSystemLoader):
    def __init__(
        self,
        namespace_name: str,
        path: Union[str, List[str]] = 'templates',
        encoding: str = 'utf-8',
        followlinks: bool = False
    ) -> None:

        if isinstance(path, str):
            path = [path]

        namespace = importlib.import_module(namespace_name)
        namespace_path = list(namespace.__path__)[0]
        searchpath = [os.path.join(namespace_path, p) for p in path]

        super().__init__(searchpath=searchpath, encoding=encoding, followlinks=followlinks)


def get_env() -> Environment:
    global _env

    if _env is None:
        logger.debug('creating Jinja2 template environment')
        loader = NamespaceLoader('qtoggleserver', [f'{FRONTEND_DIR_DEBUG}/html', f'{FRONTEND_DIR}/html'])
        _env = Environment(loader=loader, autoescape=select_autoescape())

    return _env




import abc

from typing import Callable, List, Optional, Set

from qtoggleserver.core.typing import PortValue as CorePortValue

from . import parse
from .base import Expression
from .exceptions import InvalidExpression


FUNCTIONS = {}


def function(name: str) -> Callable:
    def decorator(func_class: type) -> type:
        func_class.NAME = name
        FUNCTIONS[name] = func_class

        return func_class

    return decorator


class Function(Expression, metaclass=abc.ABCMeta):
    NAME = None
    MIN_ARGS = None
    MAX_ARGS = None

    def __init__(self, args: List[Expression]) -> None:
        self.args: List[Expression] = args
        self._deps: Optional[Set[str]] = None

    def __str__(self) -> str:
        s = getattr(self, '_str', None)
        if s is None:
            args_str = ', '.join(str(e) for e in self.args)
            self._str = s = f'{self.NAME}({args_str})'

        return s

    def get_deps(self) -> Set[str]:
        if self._deps is None:
            self._deps = set()

            for arg in self.args:
                self._deps |= arg.get_deps()

        return self._deps

    def eval_args(self) -> List[CorePortValue]:
        return [a.eval() for a in self.args]

    @staticmethod
    def parse(self_port_id: Optional[str], sexpression: str) -> Expression:
        sexpression = sexpression.strip()

        p_start = None
        p_end = None
        p_last_comma = None
        level = 0
        sargs = []
        for i, c in enumerate(sexpression):
            if c == '(':
                if p_start is None:
                    p_start = i

                elif level == 0:
                    raise InvalidExpression('Empty function call')

                level += 1

            elif c == ')':
                if level == 0:
                    raise InvalidExpression('Unbalanced parentheses')

                elif level == 1:
                    if (p_end is None) and (i == len(sexpression) - 1):
                        p_end = i

                    else:
                        raise InvalidExpression('Unexpected text after function call')

                level -= 1

            elif c == ',' and level == 1:
                sargs.append(sexpression[(p_last_comma or p_start) + 1: i])
                p_last_comma = i

        if (p_start is None) or (p_end is None) or (p_start > p_end) or (level != 0):
            raise InvalidExpression('Unbalanced parentheses')

        if p_end - p_start > 1:
            sargs.append(sexpression[(p_last_comma or p_start) + 1: p_end])

        func_name = sexpression[:p_start].strip()
        func_class = FUNCTIONS.get(func_name)
        if func_class is None:
            raise InvalidExpression(f'Unknown function "{func_name}"')

        if func_class.MIN_ARGS is not None and len(sargs) < func_class.MIN_ARGS:
            raise InvalidExpression(f'Too few arguments for function "{func_name}"')

        if func_class.MAX_ARGS is not None and len(sargs) > func_class.MAX_ARGS:
            raise InvalidExpression(f'Too many arguments for function "{func_name}"')

        args = [parse(self_port_id, sa) for sa in sargs]

        return func_class(args)


# These imports are here just because we need all modules to be imported, so that @function decorator registers all
# defined functions.

from .arithmetic import function as _
from .logic import function as _
from .bitwise import function as _
from .comparison import function as _
from .sign import function as _
from .aggregation import function as _
from .rounding import function as _
from .time import function as _
from .date import function as _
from .timeprocessing import function as _
from .various import function as _




from typing import Optional

from qtoggleserver.core.typing import PortValue as CorePortValue

from .base import Expression
from .exceptions import InvalidExpression


class Constant(Expression):
    def __init__(self, value: CorePortValue, sexpression: str) -> None:
        self.value: CorePortValue = value
        self.sexpression: str = sexpression

    def __str__(self) -> str:
        return self.sexpression

    def eval(self) -> CorePortValue:
        return self.value

    @staticmethod
    def parse(self_port_id: Optional[str], sexpression: str) -> Expression:
        sexpression = sexpression.strip()

        if sexpression == 'true':
            value = 1

        elif sexpression == 'false':
            value = 0

        else:
            try:
                value = int(sexpression)

            except ValueError:
                try:
                    value = float(sexpression)

                except ValueError:
                    raise InvalidExpression(f'"{sexpression}" is not a valid constant') from None

        return Constant(value, sexpression)




from qtoggleserver.core.typing import PortValue as CorePortValue


class ExpressionException(Exception):
    pass


class InvalidExpression(ExpressionException):
    pass


class InvalidArgument(ExpressionException):
    def __init__(self, arg_no: int, value: CorePortValue) -> None:
        self.arg_no: int = arg_no
        self.value: CorePortValue = value

    def __str__(self) -> str:
        return f'invalid argument {self.arg_no}: {self.value}'


class CircularDependency(ExpressionException):
    pass


class IncompleteExpression(ExpressionException):
    pass


class EvalSkipped(ExpressionException):
    pass




from __future__ import annotations

import abc
import logging
import time

from qtoggleserver.core import api as core_api
from qtoggleserver.core.typing import GenericJSONDict


logger = logging.getLogger(__package__)


class Event(metaclass=abc.ABCMeta):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_NONE
    TYPE = 'base-event'

    def __init__(self, timestamp: float = None) -> None:
        self._type: str = self.TYPE
        self._timestamp: float = timestamp or time.time()

    def __str__(self) -> str:
        return f'{self._type} event'

    async def to_json(self) -> GenericJSONDict:
        return {
            'type': self._type,
            'params': await self.get_params()
        }

    async def get_params(self) -> GenericJSONDict:
        return {}

    def get_type(self) -> str:
        return self._type

    def get_timestamp(self) -> float:
        return self._timestamp

    def is_duplicate(self, event: Event) -> bool:
        return False


class Handler(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def handle_event(self, event: Event) -> None:
        raise NotImplementedError()

    async def cleanup(self) -> None:
        pass




import asyncio

from typing import Callable


async def await_later(delay: float, coroutine: Callable, *args, loop: asyncio.AbstractEventLoop = None) -> None:
    await asyncio.sleep(delay, loop=loop)
    await coroutine(*args)




from typing import Optional

from .base import Expression


def parse(self_port_id: Optional[str], sexpression: str) -> Expression:
    sexpression = sexpression.strip()
    if sexpression.startswith('$'):
        return PortValue.parse(self_port_id, sexpression)

    elif sexpression.count('('):
        return Function.parse(self_port_id, sexpression)

    else:
        return Constant.parse(self_port_id, sexpression)


# Import core.ports after defining Expression, because core.ports.BasePort depends on Expression.
from qtoggleserver.core import ports as core_ports


async def check_loops(port: core_ports.BasePort, expression: Expression) -> None:
    seen_ports = {port}

    async def check_loops_rec(level: int, e: Expression) -> int:
        if isinstance(e, PortValue):
            p = e.get_port()
            if not p:
                return 0

            # A loop is detected when we stumble upon the initial port at a level deeper than 1
            if port is p and level > 1:
                return level

            # Avoid visiting the same port twice
            if p in seen_ports:
                return 0

            seen_ports.add(p)

            expr = await p.get_expression()
            if expr:
                lv = await check_loops_rec(level + 1, expr)
                if lv:
                    return lv

            return 0

        elif isinstance(e, Function):
            for arg in e.args:
                lv = await check_loops_rec(level, arg)
                if lv:
                    return lv

        return 0

    if await check_loops_rec(1, expression) > 1:
        raise CircularDependency(f'{port} is recursively referred')


from .constants import Constant
from .exceptions import (
    ExpressionException,
    InvalidExpression,
    InvalidArgument,
    CircularDependency,
    IncompleteExpression,
    EvalSkipped
)
from .functions import Function
from .portvalue import PortValue
